#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction

from .handlers.ssh import SSH
from .handlers.toml import TOML
from .password_helpers import ask_password
from .password_helpers import ask_question
from .password_helpers import gen_password


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


JID_EXT: str = "matrix-jitsi-web"


def subparser_adduser_jitsi(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        "adduser-jitsi", help="Add a new jitsi user"
    )
    parser.add_argument("user", help="The Username of the new jitsi user")
    parser.add_argument(
        "-p",
        "--passwd",
        help="The password of the new jitsi user. (If you don't enter a "
        "password, you will be asked later.)",
    )
    parser.set_defaults(func=adduser_jitsi)


def adduser_jitsi(arg: Namespace) -> int:
    """Add a User to the jitsi instance.

    It runs ``ask_password()``
    first. If ``ask_password()`` returns ``None`` it generates a password
    with ``gen_password()``. Then it gives the user a overview of the
    username, password and if the new user should be generated as admin
    (if you added the ``--admin`` argument). Next, it asks a question,
    if the entered values are correct with the ``ask_question`` function.

    If the ``ask_question`` function returns True, it continues. If not, it
    starts from the beginning.

    It runs the ``adduser`` method of the ``Ssh`` class.

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :return:          None
    """

    with TOML() as toml:
        address = (
            toml.get(("SSH", "Address"))
            if toml.get(("SSH", "Address"))
            else f"matrix.{toml.get(('API', 'Domain'))}"
        )
        with SSH(
            address, toml.get(("SSH", "User")), toml.get(("SSH", "Port"))
        ) as ssh:
            while True:
                passwd_generated: bool = False

                if arg.passwd is None:
                    arg.passwd = ask_password()

                if arg.passwd == "":
                    arg.passwd = gen_password()
                    passwd_generated = True

                print(f"Username: {arg.user}")

                if passwd_generated:
                    print(f"Password (generated): {arg.passwd}")
                else:
                    print("Password: **HIDDEN**")

                answer = ask_question()

                if answer:
                    break
                arg.passwd = None

            cmd: str = (
                "sudo docker exec matrix-jitsi-prosody prosodyctl "
                f"--config /config/prosody.cfg.lua register "
                f'"{arg.user}" {JID_EXT} "{arg.passwd}"'
            )

            ssh.run_cmd(cmd)

            return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import json
import sys

from logging import critical
from logging import debug
from types import TracebackType
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from urllib.parse import urlparse

import requests

from matrixctl import __version__
from matrixctl.errors import InternalResponseError


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def check_type(var: Any, var_name: str, var_type: Any = str) -> None:
    if not isinstance(var, var_type):
        raise TypeError(
            f"{var_name} needs to be of type {var_type}, you have entered "
            f"{type(var)}."
        )


class UrlBuilder:
    __slots__ = (
        "__scheme",
        "__subdomain",
        "__domain",
        "__api_path",
        "__api_version",
        "__path",
    )

    def __init__(self, default_domain: str) -> None:
        self.__scheme: str = "https"
        self.__subdomain: str = "matrix"
        self.__domain: str = default_domain
        self.__api_path: str = "_synapse/admin"
        self.__api_version: str = "v2"
        self.__path: str = ""

    def _scheme(self, scheme: str) -> None:
        check_type(scheme, "scheme")
        self.__scheme = scheme

    def _domain(self, domain: str) -> None:
        check_type(domain, "domain")
        self.__domain = domain

    def _subdomain(self, subdomain: str) -> None:
        check_type(subdomain, "subdomain")
        self.__subdomain = subdomain

    def _api_path(self, api_path: str) -> None:
        check_type(api_path, "api_path")
        self.__api_path = api_path

    def _api_version(self, api_version: str) -> None:
        check_type(api_version, "api_version")
        self.__api_version = api_version

    def _path(self, path: str) -> None:
        check_type(path, "path")
        self.__path = path

    scheme = property(fset=_scheme)
    domain = property(fset=_domain)
    subdomain = property(fset=_subdomain)
    api_path = property(fset=_api_path)
    api_version = property(fset=_api_version)
    path = property(fset=_path)

    def build(self) -> str:
        # Url Generation
        url: str = (
            f"{self.__scheme}://"
            f"{self.__subdomain}{'.' if self.__subdomain else ''}"
            f"{self.__domain}"
            f"/{self.__api_path}"
            f"/{self.__api_version}"
            f"/{self.__path}"
        )
        debug(f"url (unparsed): {url}")
        url = urlparse(url).geturl()
        debug(f"url   (parsed): {url}")

        return url


class API:

    """Handle the REST API connection of synapse."""

    __slots__ = (
        "token",
        "domain",
        "json_format",
        "url",
        "__success_codes",
        "session",
        "__request",
        "__method",
        "__params",
        "__headers",
    )

    # (*range(200, 208), 226)
    RESPONSE_OK: Tuple[int, ...] = (
        200,
        201,
        202,
        203,
        204,
        205,
        206,
        207,
        226,
    )

    def __init__(
        self, domain: str, token: str, json_format: bool = True
    ) -> None:
        """Initialize the Api class and checks, if the parameter are there.

        :param api_domain:  The API domain (e.g. "domain.tld")
        :param api_token:   The access token of an admin
        :return:            ``None``
        """

        check_type(domain, "domain")
        check_type(token, "token")
        check_type(json_format, "json_format", bool)

        self.token: str = token
        self.json_format: bool = json_format

        self.url: UrlBuilder = UrlBuilder(domain)
        self.__success_codes: Tuple[int, ...] = self.__class__.RESPONSE_OK

        self.session = requests.Session()

        self.__method: str = "GET"
        self.__params: Dict[str, str] = {}
        self.__headers: Dict[str, str] = {
            "User-Agent": f"matrixctl{__version__}",
            "Authorization": f"Bearer {self.token}",
        }

    def _method(self, method: str) -> None:
        method = method.upper()
        if method not in {"GET", "POST", "PUT", "DELETE"}:
            raise ValueError(
                'method needs to be one of {"GET", "POST", "PUT", "DELETE"}, '
                f"you have entered {type(method)}."
            )
        self.__method = method

    def _params(self, params: Dict[str, str]) -> None:
        check_type(params, "params", dict)
        self.__params.update(params)

    def _headers(self, headers: Dict[str, str]) -> None:
        check_type(headers, "headers", dict)

        if self.json_format:
            self.__headers["Content-Type"] = "application/json"

        self.__headers.update(headers)

    def _success_codes(self, codes: Tuple[int]) -> None:
        check_type(codes, "codes", dict)
        self.__success_codes = codes

    method = property(fset=_method)
    params = property(fset=_params)
    headers = property(fset=_headers)
    success_codes = property(fset=_success_codes)

    def request(
        self, data: Union[str, None, Dict[str, Any]] = None
    ) -> requests.Response:
        """Send a request to the synapse API.

        :param path:           The path of the request
        :param params:         Params of the request
        :param payload:        The payload of the request
        :param method:         The response method: ``GET``, ``POST``,
                               ``PUT``, ``DELETE`` of the request
        :param api_version:    The version of the api of the request
        :param success_codes:  A tuple with success return code of a requests
                               response
        :param json_payload:   ``True`` if the request and the response should
                               be in the json format.
        :return:               Returns the response
        """
        debug("Started API request.")

        url: str = self.url.build()

        if self.json_format and data is not None:
            data = json.dumps(data)

        debug_headers = self.__headers.copy()
        # len("Bearer ") = 7
        debug_headers[
            "Authorization"
        ] = f"HIDDEN (Length={len(self.__headers['Authorization'])-7})"
        debug(f"Method: {self.__method}")
        debug(f"Headers: {debug_headers}")
        debug(f"Params: {self.__params}")
        debug(f"Data: {data}")

        response = self.session.request(
            method=self.__method,
            data=data,
            url=url,
            params=self.__params,
            headers=self.__headers,
        )

        debug(f"{response.json()=}")

        if response.status_code not in self.__success_codes:
            debug(f"{response.status_code=}")
            try:
                if response.json()["errcode"] == "M_UNKNOWN_TOKEN":
                    critical(
                        "The server rejected your access-token. "
                        "Please make sure, your access-token is correct "
                        "and up-to-date. Your access-token will change every "
                        "time, you log out."
                    )
                    sys.exit(1)
            except Exception:  # pylint: disable=broad-except
                pass

            raise InternalResponseError(payload=response)

        return response

    def __enter__(self) -> API:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction

from .handlers.ssh import SSH
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


JID_EXT: str = "matrix-jitsi-web"


def subparser_deluser_jitsi(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        "deluser-jitsi", help="Deletes a jitsi user"
    )
    parser.add_argument("user", help="The jitsi username to delete")
    parser.set_defaults(func=deluser_jitsi)


def deluser_jitsi(arg: Namespace) -> int:
    """Delete a user from the jitsi instance.

    It uses the ``Ssh`` class from the ``ssh_handler``.

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :param cfg:       The ``Config`` class
    :return:          None
    """
    with TOML() as toml:
        address = (
            toml.get(("SSH", "Address"))
            if toml.get(("SSH", "Address"))
            else f"matrix.{toml.get(('API','Domain'))}"
        )
        with SSH(
            address, toml.get(("SSH", "User")), toml.get(("SSH", "Port"))
        ) as ssh:
            cmd: str = (
                "sudo docker exec matrix-jitsi-prosody prosodyctl "
                "--config /config/prosody.cfg.lua deluser "
                f'"{arg.user}@{JID_EXT}"'
            )

            ssh.run_cmd(cmd)

        return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from sys import version_info
from typing import Any
from typing import Optional

from pkg_resources import get_distribution


class Error(Exception):
    BUGMSG: str = (
        "If you discover this message, please try updating "
        "MatrixCtl. If you see this message again, we would "
        "be glad, if you would run the same command again in debug-mode "
        '(matrixctl -d [...]) and hand in a "Bug report" at '
        "https://github.com/MichaelSasser/matrixctl/issues "
        "with the complete output.\n\n"
        f"Python version: {version_info.major}.{version_info.minor}."
        f"{version_info.micro} {version_info.releaselevel}\n"
        f"MatrixCtl version: {get_distribution('matrixctl').version} \n"
    )

    def __init__(
        self, message: Optional[str] = None, payload: Any = None
    ) -> None:  # pylint: disable=keyword-arg-before-vararg
        """Use this error like a normal error in your day-to-day programming.

        This is a commandline application. Therefor no user should ever see an
        error (except in debug-mode). This error informs the user that,
        getting a traceback is a bug in this application. It gives the person
        instructions, how to hand in a bug report, to contain them asap.
        """
        self.payload: Any = payload
        msg: str = self.__class__.BUGMSG

        if message:
            msg += "\nFor developers: " + message

        super().__init__(msg)


class ConfigFileError(Error):
    pass


class InternalResponseError(Error):
    pass


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import error

from .errors import InternalResponseError
from .handlers.ansible import Ansible
from .handlers.api import API
from .handlers.toml import TOML
from .password_helpers import ask_password
from .password_helpers import ask_question
from .password_helpers import gen_password


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_adduser(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        "adduser", help="Add a new matrix user"
    )
    parser.add_argument("user", help="The Username of the new user")
    parser.add_argument(
        "-p",
        "--passwd",
        help="The password of the new user. (If you don't enter a password, "
        "you will be asked later.)",
    )
    parser.add_argument(
        "-a", "--admin", action="store_true", help="Create as admin user"
    )
    parser.add_argument(
        "--ansible", action="store_true", help="Use ansible insted of the api"
    )
    parser.set_defaults(func=adduser)


def adduser(arg: Namespace) -> int:
    """Add a User to the synapse instance.

    It runs ``ask_password()`` first. If ``ask_password()`` returns ``None``
    it generates a password with ``gen_password()``. Then it gives the user
    a overview of the username, password and if the new user should be
    generated as admin (if you added the ``--admin`` argument). Next, it asks
    a question, if the entered values are correct with the ``ask_question``
    function.

    If the ``ask_question`` function returns True, it continues. If not, it
    starts from the beginning.

    Depending on the ``--ansible`` switch it runs the ``adduser`` command
    via ansible or the API

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :return:          None
    """

    with TOML() as toml:
        with API(
            toml.get(("API", "Domain")), toml.get(("API", "Token"))
        ) as api, Ansible(toml.get(("SYNAPSE", "Path"))) as ansible:

            while True:
                passwd_generated: bool = False

                if arg.passwd is None:
                    arg.passwd = ask_password()

                if arg.passwd == "":
                    arg.passwd = gen_password()
                    passwd_generated = True

                print(f"Username: {arg.user}")

                if passwd_generated:
                    print(f"Password (generated): {arg.passwd}")
                else:
                    print("Password: **HIDDEN**")
                print(f"Admin:    {'yes' if arg.admin else 'no'}")

                answer = ask_question()

                if answer:
                    break
                arg.passwd = None

            if arg.ansible:
                arg.admin = "yes" if arg.admin else "no"
                ansible.tags = ("register-user",)
                ansible.extra_vars = {
                    "username": arg.user,
                    "password": arg.passwd,
                    "admin": arg.admin,
                }
                ansible.run_playbook()
            else:
                try:
                    api.url.path = (
                        f"users/@{arg.user}:{toml.get(('API','Domain'))}"
                    )
                    api.method = "PUT"
                    api.request({"password": arg.passwd, "admin": arg.admin})
                except InternalResponseError:
                    error("The User was not added.")

        return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import datetime
import sys

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import debug
from logging import error
from logging import fatal
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from tabulate import tabulate

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML
from .print_helpers import human_readable_bool
from .typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_user(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        "user", help="Get information about a specific user"
    )
    parser.add_argument("user", help="The username of the user")
    parser.set_defaults(func=user)


def make_human_readable(
    k: str, user_dict: Dict[str, str], len_domain: int
) -> Tuple[str, str]:

    key: Optional[str] = None
    value: Union[str]

    if k == "name":
        value = str(user_dict[k][1:-len_domain])
    elif k == "is_guest":
        value = human_readable_bool(user_dict[k])
        key = "Guest"
    elif k in ("admin", "deactivated"):
        value = human_readable_bool(user_dict[k])
    elif k.endswith("_ts"):
        value = str(
            datetime.datetime.fromtimestamp(float(user_dict[k]))
        )  # UTC?
    elif k.endswith("_at"):
        value = str(
            datetime.datetime.fromtimestamp(float(user_dict[k]) / 1000.0)
        )  # UTC?

    else:
        value = user_dict[k]

    if key is None:
        key = k.replace("_", " ").title()

    return key, value


def generate_user_tables(
    user_dict: Dict[str, Any], len_domain: int
) -> List[List[Tuple[str, str]]]:
    """Generate a main user table and threepid user tables.

    The function gnerates first a main user table and then for every threepid
    a additional table from a ``user_dict``.
    It renames and makes the output human readable.

    This function is a recursive function

    :param user_dict:   A JSON string which was converted to a Python
                        dictionary.
    :param len_domain:  The length in characters of the domain.
    :return:            The generated lists in this format:
                        [main], threepids_0, ... ,threepids_n]
    """

    table: List[List[Tuple[str, str]]] = [[]]

    for k in user_dict:
        if k == "errcode":
            error("There is no user with that username.")
            sys.exit(1)

        if k == "threepids":
            for tk in user_dict[k]:
                ret: List[List[Tuple[str, str]]] = generate_user_tables(
                    tk, len_domain
                )
                table.append(ret[0])

            continue  # Don't add threepids to "table"

        table[0].append(make_human_readable(k, user_dict, len_domain))

    return table


def user(arg: Namespace) -> int:
    """List information about an registered user.

    It uses the admin API to get a python dictionary with the information.
    The ``generate_user_tables`` function makes the information human readable.
    The Python package ``tabulate`` renders the table as shown below, if
    everything works well.


    .. code-block:: console

       $ matrixctl user dwight
       User:
       +----------------------------+--------------------------------------+
       | Name                       | dwight                               |
       | Password Hash              | $2b$12$9DUNderm1ffL1NincPap3RC       |
       |                            | ompaNY1725.slOUghAvEnu5cranT0n       |
       | Guest                      | False                                |
       | Admin                      | True                                 |
       | Consent Version            |                                      |
       | Consent Server Notice Sent |                                      |
       | Appservice Id              |                                      |
       | Creation Ts                | 2020-04-14 13:04:21                  |
       | User Type                  |                                      |
       | Deactivated                | False                                |
       | Displayname                | Dwight Schrute                       |
       | Avatar Url                 | mxc://dunder-mifflin.com/sCr4        |
       |                            | nt0nsr4ng13rW45Cr33d                 |
       +----------------------------+--------------------------------------+

       Threepid:
       +--------------+-----------------------------------+
       | Medium       | email                             |
       | Address      | dwight_schrute@dunder-mifflin.com |
       | Validated At | 2020-04-14 15:30:21.123000        |
       | Added At     | 2020-04-14 15:29:19.100000        |
       +--------------+-----------------------------------+


    If the user does not exist, the return looks like:


    .. code-block:: console

       $ matrixctl user mose
       2020-04-14 13:58:13 - ERROR - The request was not successful.
       2020-04-14 13:58:13 - ERROR - There is no user with that username.

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :param cfg:       The ``Config`` class
    :return:          None
    """

    with TOML() as toml:
        with API(
            toml.get(("API", "Domain")), toml.get(("API", "Token"))
        ) as api:
            api.url.path = f'users/@{arg.user}:{toml.get(("API","Domain"))}'

            try:
                user_dict: JsonDict = api.request().json()
            except InternalResponseError:
                fatal("Could not receive the user information")

                return 1

            len_domain = len(toml.get(("API", "Domain"))) + 1  # 1 for :
            user_tables = generate_user_tables(user_dict, len_domain)

            debug(f"User: {user_tables=}")

            for num, table in enumerate(user_tables):

                if num < 1:
                    print("User:")
                else:
                    print("\nThreepid:")
                print(tabulate(table, tablefmt="psql",))

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import sys
import warnings

from copy import deepcopy
from logging import debug
from logging import error
from pathlib import Path
from types import TracebackType
from typing import Any
from typing import Dict
from typing import ItemsView
from typing import Iterator
from typing import KeysView
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import ValuesView

import toml

from matrixctl.errors import ConfigFileError


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


class TOML:
    def __init__(self, path: Optional[Path] = None) -> None:
        debug("Loading Config file(s)")

        if path is None:
            self.__default: bool = True

        default_config: List[str] = [
            "/etc/matrixctl/config",
            str(Path.home() / ".config/matrixctl/config"),
        ]
        # self.__check_paths(self.__class__.FILE_PATH)

        # with warnings.catch_warnings():
        #     captureWarnings(True)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                # toml.load is a MutableMapping
                self.__toml: Dict[str, Any] = deepcopy(
                    dict(
                        toml.load(
                            default_config if self.__default else str(path)
                        )
                    )
                )
            except FileNotFoundError:
                error(
                    "To use this program you need to have a config file in"
                    '/etc/matrixctl/config" or in '
                    '"~/.config/matrixctl/config".'
                )
                sys.exit(1)
            except TypeError:
                raise ConfigFileError()
            except toml.TomlDecodeError:
                error(
                    "Please check your config file. MatrixCtl was not able "
                    "to read it."
                )
                sys.exit(1)

        self.__debug_output()

    def __debug_output(self) -> None:
        for key in self.__toml:
            debug(f"[{key}]")

            for entry in self.__toml[key]:
                if entry == "Token":
                    length = len(self.__toml[key][entry])
                    debug(f"  ├─  {entry} := **HIDDEN (Length={length})**")
                else:
                    debug(f"  ├─  {entry} := {self.__toml[key][entry]}")
            debug("  ┴")

    def get(self, keys: Union[List[str], Tuple[str, ...]]) -> Any:
        """Get a value from a config entry safely.

        This is the only way, the config is asked for a value from an entry.
        The other methods are used for non MatrixCtl TOML files.

        **Usage**

        Pass in a list or tuple with strings describing the path in the
        ``self.__toml`` dictionary.
        Let's say, you are looking for the synapse path:

        >>> toml.get(("SYNAPSE", "Path"))
        '/home/dwight/SomRandomDirectory/synapse'

        :param keys:  A string or tuple describing the values you are looking
                      for.
        :return:      The value of the entry you described.
        """
        toml_walker: Union[Dict[str, Any], Any] = self.__toml

        try:
            for key in keys:
                toml_walker = toml_walker.__getitem__(key)
        except KeyError:
            error(
                "Please check your config file. For this operation your "
                f'config file needs to have the entry "{keys[-1]}" '
                f'In the section "[{keys[0]}]".'
            )
            sys.exit(1)

        if not isinstance(toml_walker, dict):
            return toml_walker
        raise ConfigFileError(
            "The key you asked for seems not correct. "
            "Please make sure you ask for an single entry, "
            "not a entire section."
        )

    def has_key(self, key: str) -> bool:
        return key in self.__toml

    def keys(self) -> KeysView[str]:
        return self.__toml.keys()

    def values(self) -> ValuesView[Any]:
        return self.__toml.values()

    def items(self) -> ItemsView[str, Any]:
        return self.__toml.items()

    def copy(self) -> Dict[str, str]:
        return self.__toml.copy()

    def __getitem__(self, key: str) -> Any:
        return self.__toml[key]

    def __len__(self) -> int:
        return len(self.__toml)

    def __contains__(self, item: str) -> bool:
        return item in self.__toml

    def __iter__(self) -> Iterator[str]:
        return iter(self.__toml)

    def __enter__(self) -> TOML:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return

    def __repr__(self) -> str:
        return repr(self.__toml)

    def __str__(self) -> str:
        return str(self.__toml)


#
# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any
from typing import Dict


JsonDict = Dict[str, Any]

# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import fatal
from typing import List
from typing import Tuple

from tabulate import tabulate

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML
from .print_helpers import human_readable_bool
from .typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_users(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser("users", help="Lists users")
    parser.add_argument(
        "-g", "--with-guests", action="store_true", help="Shows guests"
    )
    parser.add_argument(
        "-b", "--with-bots", action="store_true", help="Shows bots"
    )
    parser.add_argument(
        "-d",
        "--with-deactivated",
        action="store_true",
        help="Shows deactivated accounts",
    )
    parser.set_defaults(func=users)


def users(arg: Namespace) -> int:
    """Print a table of the matrix users.

    This function generates and prints a table of matrix user accounts.
    The table can be modified with:

    - the ``--guests`` switch: ``args.guests`` is True, else ``False``
    - the ``--no-bots`` switch: ``args.no_bots`` is True, else ``False``

    If you don't want any bots in the table use the ``--no-bots`` switch.
    If you want guests in the table use the ``--guests`` switch.

    **Example**

    .. code-block:: console

       $ matrixctl users --no-bots
       +----------------+---------------+------------+------------+
       | Name           | Deactivated   | Is Admin   | Is Guest   |
       |----------------+---------------+------------+------------|
       | dunder_mifflin | False         | True       | False      |
       | dwight         | False         | True       | False      |
       | pam            | False         | False      | False      |
       | jim            | False         | False      | False      |
       | creed          | False         | False      | False      |
       | stanley        | False         | False      | False      |
       | kevin          | False         | False      | False      |
       | angela         | False         | False      | False      |
       | phyllis        | False         | False      | False      |
       | tobi           | False         | False      | False      |
       | michael        | False         | True       | False      |
       | andy           | False         | False      | False      |
       +----------------+---------------+------------+------------+

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :return:          None
    """
    with TOML() as toml:
        len_domain = len(toml.get(("API", "Domain"))) + 1  # 1 for :
        from_user: int = 0
        users_list: List[JsonDict] = []

        # ToDo: API bool
        with API(
            toml.get(("API", "Domain")), toml.get(("API", "Token"))
        ) as api:
            api.url.path = "users"
            api.params = {"guests": "true" if arg.with_guests else "false"}
            api.params = {
                "deactivated": "true" if arg.with_deactivated else "false"
            }

            while True:

                api.params = {"from": from_user}  # from must be in the loop
                try:
                    lst: JsonDict = api.request().json()
                except InternalResponseError:
                    fatal("Could not get the user table.")

                    return 1

                users_list += lst["users"]

                try:
                    from_user = lst["next_token"]
                except KeyError:
                    break

            user_list: List[Tuple[str, str, str, str]] = []

            for user in users_list:
                name = user["name"][1:-len_domain]
                no_passwd_hash: bool = user["password_hash"] == ""
                deactivated: str = human_readable_bool(user["deactivated"])
                admin: str = human_readable_bool(user["admin"])
                guest: str = human_readable_bool(user["is_guest"])

                # if no_bots and any([name.startswith(bot) for bot in BOTS]):
                #     continue

                if not arg.with_bots and no_passwd_hash:
                    continue

                user_list.append((name, deactivated, admin, guest,))
            print(
                tabulate(
                    user_list,
                    headers=("Name", "Deactivated", "Is Admin", "Is Guest"),
                    tablefmt="psql",
                )
            )

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import fatal
from typing import List
from typing import Tuple

from tabulate import tabulate

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML
from .typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_rooms(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser("rooms", help="List rooms")
    parser.add_argument(
        "-s",
        "--order-by-size",
        action="store_true",
        help="Order the rooms by size instead of alphabetical",
    )
    parser.add_argument(
        "-r", "--reverse", action="store_true", help="Reverse the order"
    )
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        default=0,
        help="The number of rooms to show",
    )
    parser.add_argument(
        "filter",
        type=str,
        nargs="?",
        default=None,
        help="Search for rooms with a specific search term (Case-Sensitive)",
    )
    parser.set_defaults(func=rooms)


def rooms(arg: Namespace) -> int:
    """Generate a table of the matrix rooms.

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :return:          None
    """
    with TOML() as toml:
        from_room: int = 0
        rooms_list: List[JsonDict] = []

        with API(
            toml.get(("API", "Domain")), toml.get(("API", "Token"))
        ) as api:
            api.url.path = "rooms"
            api.url.api_version = "v1"

            if arg.number > 0:
                api.params = {"limit": arg.number}

            if arg.filter:
                api.params = {"search_term": arg.filter}

            if arg.reverse:
                api.params = {"dir": "b"}

            if arg.order_by_size:
                api.params = {"order_by": "size"}

            while True:

                api.params = {"from": from_room}  # from must be in the loop
                try:
                    lst: JsonDict = api.request().json()
                except InternalResponseError:
                    fatal("Could not get the room table.")

                    return 1

                rooms_list += lst["rooms"]
                try:
                    from_room = lst["next_token"]
                except KeyError:
                    break
    print_rooms_table(rooms_list)

    return 0


def print_rooms_table(rooms_list: List[JsonDict]) -> None:

    room_list: List[Tuple[str, int, str, str]] = []

    for room in rooms_list:
        name = room["name"]
        members: int = room["joined_members"]
        alias: str = room["canonical_alias"]
        room_id: str = room["room_id"]

        room_list.append((name, members, alias, room_id,))
    print(
        tabulate(
            room_list,
            headers=("Name", "Members", "Alias", "Room ID"),
            tablefmt="psql",
        )
    )


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# import configparser
from __future__ import annotations

import datetime
import sys

from logging import debug
from logging import error
from logging import info
from pathlib import Path
from shutil import get_terminal_size
from textwrap import TextWrapper
from types import TracebackType
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import git

from tabulate import tabulate


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


class Git:
    def __init__(self, path: Union[Path, str]) -> None:
        self.path: Path = Path(path)
        self.repo = git.Repo(self.path)
        assert not self.repo.bare

        self.git = git.cmd.Git(self.path)
        self.heads = self.repo.heads
        self.master = self.heads.master

    @property
    def datetime_last_pulled_commit(self) -> datetime.datetime:
        log = self.master.log()

        return datetime.datetime.fromtimestamp(log[-1].time[0])

    def log(self, since: Optional[datetime.datetime] = None) -> None:
        cmd = ["--pretty=%as\t%an\t%s"]

        if since:
            cmd.append(f"--since={str(since)}")

        terminal_size_x, _ = get_terminal_size()
        debug(f"Terminal width = {terminal_size_x}")

        ######################################################################
        #                          Terminal width                            #
        #                          ^^^^^^^^^^^^^^                            #
        #                                                                    #
        # |<-------------------------------(121)-------------------------->| #
        # |<--------------------------(119)------------------------->|     | #
        # |<----------------------(34)--------------------->|        |     | #
        # |<------------------(32)----------------->|       |        |     | #
        # |<---------------(30)-------------->|     |       |        |     | #
        # |<-------------16------------>|     |     |       |        |     | #
        # |<---------14--------->|      |     |     |       |        |     | #
        # |<----12----->|        |      |     |     |       |        |     | #
        # |<--3-->|     |        |      |     |     |       |        |     | #
        # |_______d a t e________|______u s e r_____|_______commit_msg_____| #
        #                               |<--->|            |<------>|        #
        #                                 15                 x - 35          #
        #                                                                    #
        ######################################################################

        wrapper_user = TextWrapper(
            width=15, drop_whitespace=True, break_long_words=True
        )
        wrapper_comment = TextWrapper(
            width=terminal_size_x - 35,
            drop_whitespace=True,
            break_long_words=True,
        )

        log: List[List[str]] = [
            line.split("\t") for line in self.git.log(cmd).split("\n")
        ]

        if not log[0][0]:  # Nothing new
            info("Everything is up-to-date.")

            return

        for line in log:
            line[1] = wrapper_user.fill(text=line[1])
            line[2] = wrapper_comment.fill(text=line[2])

        print(
            tabulate(
                log,
                headers=("Date", "User", "Commit Message"),
                tablefmt="psql",
            )
        )

    def pull(self) -> None:
        # Get the last pulled datetime
        since = self.datetime_last_pulled_commit

        try:
            self.git.pull()
        except git.GitCommandError:
            error(
                "MatrixCtl was not able to connect to the synapse playbook "
                "on GitHub. Are you connected to the internet?"
            )
            sys.exit(1)

        self.log(since)

    def __enter__(self) -> Git:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import error

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_deluser(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        "deluser", help="Deletes a user"
    )
    parser.add_argument("user", help="The username to delete")
    parser.set_defaults(func=deluser)


def deluser(arg: Namespace) -> int:
    """Delete a user from the the matrix instance.

    It uses the synapse admin API.

    :param arg:       The ``Namespace`` object of argparse's ``arse_args()``
    :param _:         Not used (The ``Config`` class)
    :return:          None
    """
    with TOML() as toml:
        with API(
            toml.get(("API", "Domain")), toml.get(("API", "Token"))
        ) as api:
            try:
                api.url.path = (
                    f"deactivate/@{arg.user}:{toml.get(('API','Domain'))}"
                )
                api.url.api_version = "v1"
                api.method = "POST"
                api.request({"erase": True})
            except InternalResponseError:
                error("The user was not deleted.")

        return 0


# vim: set ft=python :



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['matrixctl', 'matrixctl.handlers']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.0,<4.0.0',
 'argcomplete>=1.11.1,<2.0.0',
 'coloredlogs>=14.0,<15.0',
 'paramiko>=2.7.1,<3.0.0',
 'requests>=2.23.0,<3.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'toml>=0.10.0,<0.11.0']

extras_require = \
{'docs': ['sphinx>=3.0,<4.0',
          'sphinx_rtd_theme>=0.4.3,<0.5.0',
          'sphinx-autodoc-typehints>=1.10,<2.0',
          'sphinxcontrib-programoutput>=0.16,<0.17']}

entry_points = \
{'console_scripts': ['matrixctl = matrixctl.application:main']}

setup_kwargs = {
    'name': 'matrixctl',
    'version': '0.6.1',
    'description': 'Controls a synapse oci-container instance via ansible',
    'long_description': '![GitHub](https://img.shields.io/github/license/MichaelSasser/matrixctl?style=flat-square)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/matrixctl?style=flat-square)\n![PyPI - Wheel](https://img.shields.io/pypi/wheel/matrixctl?style=flat-square)\n![GitHub release (latest SemVer)](https://img.shields.io/github/v/release/michaelsasser/matrixctl?style=flat-square)\n![GitHub Release Date](https://img.shields.io/github/release-date/michaelsasser/matrixctl?style=flat-square)\n![PyPI - Status](https://img.shields.io/pypi/status/matrixctl?style=flat-square)\n![Matrix](https://img.shields.io/matrix/matrixctl:michaelsasser.org?server_fqdn=matrix.org&style=flat-square)\n![GitHub code size in bytes](https://img.shields.io/github/languages/code-size/michaelsasser/matrixctl?style=flat-square)\n\n# MatrixCtl\n\nMatrixCtl is a python program to control, manage, provision and deploy our\nmatrix homeserver. Instead of remembering tons of commands or having a bunch\nof shell scripts MatrixCtl does many things for you.\n\n## Command line tool\n\nMatrixCtl as a pure commandline tool. You can use it as package, if you like,\nbut breaking changes may be introduced, even in a minor version shift.\n\n```\n# matrixctl\nusage: matrixctl [-h] [--version] [-d]\n              {adduser,deluser,adduser-jitsi,deluser-jitsi,user,users,rooms,delroom,update,deploy,start,restart,maintenance,check,version}\n              ...\n\npositional arguments:\n  {adduser,deluser,adduser-jitsi,deluser-jitsi,user,users,rooms,delroom,update,deploy,start,restart,maintenance,check,version}\n    adduser             Add a new matrix user\n    deluser             Deletes a user\n    adduser-jitsi       Add a new jitsi user\n    deluser-jitsi       Deletes a jitsi user\n    user                Get information about a specific user\n    users               Lists users\n    rooms               List rooms\n    delroom             Deletes an empty room from the database\n    update              Updates the ansible repo\n    deploy              Provision and deploy\n    start               Starts all OCI containers\n    restart             Restarts all OCI containers (alias for start)\n    maintenance         Run maintenance tasks\n    check               Checks the OCI containers\n    version             Get the version of the Synapse instance\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --version             show program\'s version number and exit\n  -d, --debug           Enables debugging mode.\n```\n\n## Installation\n\nMatrixCtl is written in Python. The installation is straight forward. Just run ``pip install matrixctl``. MatrixCtl will be installd from the [Python Package Index (PyPi)](https://pypi.org/project/matrixctl/).\n\nYou will find more information in the documentation.\n\n## Documentation\n\nThere is a [documentation](https://michaelsasser.github.io/matrixctl/index.html) waiting for you, showing you how everything works and howto setup matrixctl\n\n## Configuration File\n\nTo use this program you need to have this config file in\n"/etc/matrixctl/config" or in "~/.config/matrixctl/config".\n\n```toml\n[SYNAPSE]\n# The absolute path to the fully configured matrix-docker-ansible-deploy\n# playbook from https://github.com/spantaleev/matrix-docker-ansible-deploy.\n#\n# Path="/absolut/path/to/matrix-docker-ansible-deploy"\n\n[ANSIBLE]\n# If you have your own playbook, to provision your matrix server, you can\n# fill out this section. MatrixCtl will run this before the synapse playbook.\n\n# The absolute path to your playbook\n#\n# Path = "/absolute/path/to/your/playbook"\n\n# If you have a special "ansible.cfg" for your playbook, fill in the absolute\n# path to it.\n#\n# Cfg="/absolute/path/to/ansible.cfg"\n\n# Fill in the absolute path to your Playbook (e.g. "site.yml")\n#\n# Playbook ="setup.yml"\n\n# If you use tags to provision or configure your matrix host, you can add them\n# here.\n#\n# Tags = ["MyTag", "MyOtherTag"]\n\n[API]\n# If your matrix server is deployed, you may want to fill out the API section.\n# It enables matrixctl to run more and faster commands. You can deploy and\n# provision your Server without this section. You also can cerate a user with\n# "matrixctl adduser --ansible YourUsername" and add your privileges after\n# that.\n\n# Your domain should be something like "michaelsasser.org" without the\n# "matrix." in the front. MatrixCtl will add that, if needed. An IP-Address\n# is not enough.\n#\n# Domain="domain.tld"\n\n# To use the API you need to have an administrator account. Enter your Token\n# here. If you use the riot client you will find it your user settings (click\n# on your username on the upper left corner on your browser) in the\n# "Help & About" tab. If you scroll down click next to "Access-Token:" on\n# "<click to reveal>". It will be marked for you. Copy it in here.\n#\n# Token="MySuperLongMatrixToken"\n\n[SSH]\n# Here you can add your SSH configuration.\n#\n# Address = "matrix.domain.tld"\n\n# The default port is 22\n#\n# Port = 22\n\n# The default username is your current login name.\n#\n# User = "myusername"\n```\n\n## Chat\n\nIf you have any thoughts or questions, you can join the project channel ``#matrixctl:michaelsasser.org``.\n\n## Semantic Versioning\n\nThis repository uses [SemVer](https://semver.org/) for its release\ncycle.\n\n## Branching Model\n\nThis repository uses the\n[git-flow](https://danielkummer.github.io/git-flow-cheatsheet/index.html)\nbranching model by [Vincent Driessen](https://nvie.com/about/).\nIt has two branches with infinite lifetime:\n\n* [master](https://github.com/MichaelSasser/matrixctl/tree/master)\n* [develop](https://github.com/MichaelSasser/matrixctl/tree/develop)\n\nThe master branch gets updated on every release. The develop branch is the\nmerging branch.\n\n## License\nCopyright &copy; 2020 Michael Sasser <Info@MichaelSasser.org>. Released under\nthe GPLv3 license.\n',
    'author': 'Michael Sasser',
    'author_email': 'Michael@MichaelSasser.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://michaelsasser.github.io/matrixctl/index.html',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import getpass
import secrets
import string

from typing import Optional


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


SPECIAL = "!§$%&/()=?.,;:_-#+*~{}[]°^@<>|\\"
ALPHABET = string.ascii_letters + string.digits + SPECIAL


def ask_password() -> Optional[str]:
    """Ask the user to create a password.

    The user will be asked twice, after
    that, the function compares the two entered passwords. If they are the
    same, the function will return the password.

    .. note:: If the user presses enter twice, without entering a password,
              the function will return ``None``. This is by-design.

    :return: Returns the user entered password or ``None``
    """
    # ToDo: Check password (regex)
    passwd: str = "a"
    passwd2: str = "b"  # Something invalid: a==b is not True

    while passwd != passwd2:
        passwd = getpass.getpass()
        passwd2 = getpass.getpass("Password (again): ")

    return passwd if passwd == passwd2 else None


def gen_password(
    length: int = 16,
    min_digits: int = 3,
    min_special: int = 1,
    max_special: int = 3,
) -> str:
    """Generate a password.

    It uses lower- and uppercase
    characters, digits and some special characters, you can find in the
    ``SPECIAL`` variable in this file.

    It uses at least:

    - ``min_digits`` digits
    - ``min_special`` special characters

    It uses maximal:

    - ``max_special`` special characters

    The rest are lower- and uppercase characters.

    :param length:       The desired length of the password to generate
    :param min_digits:   The desired minimum of digits of the password
    :param min_special:  The desired minimum of special characters of the
                         password
    :param max_special:  The desired maximum of special characters of the
                         password
    :return:             It returns the generated password
    """

    while True:
        password = "".join(secrets.choice(ALPHABET) for i in range(length))

        # pylint: disable=chained-comparison

        if (
            any(c.islower() for c in password)
            and any(c.isupper() for c in password)
            and sum(c.isdigit() for c in password) >= min_digits
            and min_special
            <= sum(c in SPECIAL for c in password)
            <= max_special
        ):
            return password


def ask_question(question: str = "Is everything correct?") -> bool:
    """Asks the user a simple yes/no a question.

    The answer will be returned as ``bool``.
    - ``True``: if the user answered ``y`` / ``j`` (j for the German yes: Ja)
    - ``False``: if the user answers ``n``

    If the user answerd with an invalid answer (not ``y`` / ``j`` / ``n``)
    the function asks again.

    :param question:  The yes/no question the user should be asked
    :return:          ``True`` if the answer was ``y`` / ``j`` and ``False``
                      if the answer was ``n``
    """
    question += " [y/n]"

    while (answer := input(question).lower()) not in ("y", "j", "n",):  # noqa
        pass

    return answer in ("y", "j")


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import debug

from .handlers.ansible import Ansible
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_maintenance(subparsers: SubParsersAction) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        "maintenance", help="Run maintenance tasks"
    )
    parser.set_defaults(func=maintenance)


def maintenance(_: Namespace) -> int:
    debug("maintenance")

    with TOML() as toml:
        with Ansible(toml.get(("SYNAPSE", "Path"))) as ansible:
            ansible.tags = (
                "run-postgres-vacuum",
                "start",
            )
            ansible.run_playbook()

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import json
import os
import subprocess

from logging import debug
from pathlib import Path
from types import TracebackType
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from matrixctl.typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


# ToDO: __slots__
class Ansible:
    def __init__(
        self, path: Union[Path, str], playbook: str = "setup.yml"
    ) -> None:
        self.path: Path = Path(path)

        assert isinstance(playbook, str)

        self.playbook: str = playbook
        self.__ansible_cfg_path: Optional[str] = None
        self.__cmd: Dict[str, str] = {  # Defaults are already in here
            "inventory": str(self.path / "inventory/hosts"),
        }

    def _inventory_path(self, path: Path) -> None:
        assert isinstance(path, Path)
        self.__cmd["inventory"] = str(path)

    def _tags(self, tags: Iterable[str]) -> None:
        assert isinstance(tags, Iterable)
        self.__cmd["tags"] = ",".join(tags)  # e.g. tag1,tag2,...,tagn

    def _extra_vars(self, extra_vars: JsonDict) -> None:
        self.__cmd["extra-vars"] = json.dumps(extra_vars).replace(" ", "")

    def _ansible_cfg_path(self, path: Path) -> None:
        assert isinstance(path, str)
        self.__ansible_cfg_path = path

    inventory_path = property(fset=_inventory_path)
    tags = property(fset=_tags)
    extra_vars = property(fset=_extra_vars)
    ansible_cfg_path = property(fset=_ansible_cfg_path)

    def run_playbook(self) -> None:
        # Build the command
        cmd: List[str] = ["ansible-playbook"]

        for k in self.__cmd:
            cmd.append(f"--{k}={self.__cmd[k]}")

        cmd.append(str(self.path / self.playbook))

        # Debug output of the command
        debug(f"Ansible command:           {cmd}")
        debug(f"Ansible assembled command: {' '.join(cmd)}")

        # Add ansible.cfg to ENV

        if self.__ansible_cfg_path is not None:
            os.environ["ANSIBLE_CONFIG"] = self.__ansible_cfg_path

        # Run command
        subprocess.run(cmd, check=True)

        # Delete ansible.cfg from ENV

        if self.__ansible_cfg_path is not None:
            del os.environ["ANSIBLE_CONFIG"]

    def __enter__(self) -> Ansible:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Use the class with the ``with`` statement`` statement.

        This is currently not really needed, but unifies the way handlers are
        used.
        """

        return


# vim: set ft=python :
